//
//	Modified Object-oriented Realistic Image Synthesis
//	original coded C language in 1989
//	ported to C++ in 2001
//	(c) Copyright & Programmed 2001,2003,2005 by NAKATA H.
//		webmaster@nakata-jp.org
//	this code is distributed under GPL
//

#include "vector.h"
#include "color.h"
#include "pixel.h"
#include "quadratic2D.h"
#include "quadratic3D.h"
#include "primitive.h"
#include "light.h"
#include "camera.h"
#include "screen.h"
#include "andor.h"
#include "cpg.h"

#include <math.h>
#include <string.h>
#include <sstream>
#include <fstream>

#define	DEG2RAD(deg)	(deg * M_PI / 180)

static void	read_sldprim( std::ifstream &csrc );
static void	read_andor( std::ifstream &csrc );

camera		cam;
screen		scrn;
light		light0;
double		highlight;
int		prim_count, and_count, or_count;
primitive	primitives[MAXPRIMITIVES];
andclass	and_array[MAXPRIMITIVES];
orclass		or_array[MAXPRIMITIVES];

vector		cam_rot;

static char	rcsstring[] ="$Header: /home/nakata/programs/cg/src/RCS/read.cpp,v 1.16 2005/04/09 13:18:06 nakata Exp $";

void	read_sld( char *file_name )
{
  std::ifstream	csrc;
  double	tmp[5];
  char		c;

  scrn.assign( 480, 320 );

  csrc.open( file_name );
  if ( csrc.bad() ) {
    std::cerr << "source file open error.\n";
    exit(-1);
  }

  // 1st line camera definition
  csrc >> tmp[0] >> c >> tmp[1] >> c >> tmp[2] >> c >> tmp[3] >> c >> tmp[4];
  cam.point.assign( tmp[0], tmp[1], tmp[2] );
  cam_rot.x = -DEG2RAD(tmp[3]);
  cam_rot.y = -DEG2RAD(tmp[4]);
  cam.orient.assign( 0.0, 0.0, SCREENDEPTH );
  cam.horizontal.assign( SCREENWIDTH, 0.0, 0.0);
  cam.vertical.assign( 0.0, (-SCREENWIDTH * scrn.height / scrn.width), 0.0);
  cam.rotx( cam_rot.x );
  cam.roty( cam_rot.y );
  cam.point -= cam.orient;

  // 2nd line light definition
  csrc >> tmp[0] >> c >> tmp[1] >> c >> tmp[2];
  light0.assign( 0, 0, 1.0 );
  light0.rotx(-DEG2RAD(tmp[1]));
  light0.roty(-DEG2RAD(tmp[2]));

  // 3rd line highlight intensity
  csrc >> highlight;
  highlight /= MAXCHAR;

  read_sldprim( csrc );

  read_andor( csrc );

  csrc.close();
  if ( csrc.bad() ) {
    std::cerr << "source file close error.\n";
  }
}

static void	read_sldprim( std::ifstream &csrc )
{
  primitive	*p_prim;
  quadratic3D	*p_quad;
  cube		*p_cube;
  vector	tmpvec;
  double	tmp[10];
  int		i, form, rot;
  char		c;

  // primitive definitions
  for ( prim_count = 0; ; prim_count++ ) {
    if ( prim_count >= MAXPRIMITIVES ) {
      std::cerr << "too many primitives\n";
      exit(-1);
    }

    p_prim = &(primitives[ prim_count ]);

    csrc >> i;
    if ( i < 0 ) {
      return;			// end of primitives
    }
    switch ( i ) {
    case 0:
      p_prim->texture = NOTHING;
      break;
    case 1:
      p_prim->texture = CHECKER;
      break;
    case 2:
      p_prim->texture = STRIPE;
      break;
    case 3:
      p_prim->texture = CIRCLE;
      break;
    case 4:
      p_prim->texture = SPOTS;
      break;
    default:
      std::cerr << "unknown texture\n";
      exit(-1);
    }

    csrc >> c >> form >> c;
    csrc >> i >> c;
    switch ( i ) {
    case 1:
      p_prim->reflection = 0.0;
      p_prim->refraction = 0.0;
      break;
    case 2:
      p_prim->reflection = 0.95; // 5% energy loss at refrection
      p_prim->refraction = 0.0;
      break;
    }

    csrc >> rot >> c;
    csrc >> tmp[0] >> c >> tmp[1] >> c >> tmp[2] >> c;
    csrc >> tmp[3] >> c >> tmp[4] >> c >> tmp[5] >> c;
    switch ( form ) {
    case 1:
      p_cube = new cube(tmp[0], tmp[1], tmp[2], tmp[3], tmp[4], tmp[5] );
      p_prim->form3D.p_cube = p_cube;
      p_prim->cube_flag = TRUE;
      break;
    case 2:
      p_quad = new quadratic3D( 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
				tmp[0], tmp[1], tmp[2], 0 );
      p_prim->form3D.p_quad = p_quad;
      p_prim->cube_flag = FALSE;
      break;
    case 3:
      if ( tmp[0] != 0 ) {
	tmp[0] = (tmp[0] > 0) ? (1/(tmp[0]*tmp[0])) : (-1/(tmp[0]*tmp[0]));
      }
      if ( tmp[1] != 0 ) {
	tmp[1] = (tmp[1] > 0) ? (1/(tmp[1]*tmp[1])) : (-1/(tmp[1]*tmp[1]));
      }
      if ( tmp[2] != 0 ) {
	tmp[2] = (tmp[2] > 0) ? (1/(tmp[2]*tmp[2])) : (-1/(tmp[2]*tmp[2]));
      }
      p_quad = new quadratic3D( tmp[0], tmp[1], tmp[2],
				0.0, 0.0, 0.0, 0.0, 0.0, 0.0, -1.0 );
      p_prim->form3D.p_quad = p_quad;
      p_prim->cube_flag = FALSE;
      break;
    case 4:
      p_quad = new quadratic3D( tmp[0], tmp[1], tmp[2],
				0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 );
      p_prim->form3D.p_quad = p_quad;
      p_prim->cube_flag = FALSE;
      break;
    }

    csrc >> i >> c;
    p_prim->side = (i > 0) ? INNER : OUTER;

    csrc >> p_prim->diffuse >> c;
    csrc >> i >> c;
    p_prim->highlight = (double)i / 255.0;
    csrc >> i >> c;
    p_prim->col.red = (double)i / 255.0;
    csrc >> i >> c;
    p_prim->col.green = (double)i / 255.0;
    csrc >> i;
    p_prim->col.blue = (double)i / 255.0;

    if ( rot > 0 ) {
      csrc >> c >> tmp[6] >> c >> tmp[7] >> c >> tmp[8];
      if ( p_prim->cube_flag ) {
	std::cerr << "cannot rotate cube form\n";
	exit(-1);
      }

      tmpvec.assign( DEG2RAD(tmp[6]), DEG2RAD(tmp[7]), DEG2RAD(tmp[8]) );
      p_prim->form3D.p_quad->rotate( tmpvec );
    }

    if ( !p_prim->cube_flag ) {
      tmpvec.assign( tmp[3], tmp[4], tmp[5] );
      p_prim->form3D.p_quad->shift( tmpvec );
    }

  }
}

static void	read_andor( std::ifstream &csrc )
{
  andclass	*p_and;
  orclass	*p_or;
  int		i;
  char		c;

  // and definition
  for ( and_count = 0; ; and_count++ ) {
    if ( and_count >= MAXPRIMITIVES ) {
      std::cerr << "too many AND definitions.\n";
      exit(-1);
    }

    p_and = &(and_array[and_count]);
    csrc >> i;
    if ( i < 0 ) {
      break;
    } else if ( i >= prim_count ) {
      std::cerr << "illegal primitive used in AND section.\n";
      exit(-1);
    }

    p_and->length = 1;
    p_and->p_prim[0] = &(primitives[i]);

    while ( csrc >> c >> i, (i >= 0) ) {
      if ( i >= prim_count ) {
	std::cerr << "illegal primitive used in AND section.\n";
	exit(-1);
      }
      p_and->p_prim[p_and->length++] = &(primitives[i]);
    }
  }

  // or definition
  for ( or_count = 0; ; or_count++ ) {
    if ( or_count >= MAXPRIMITIVES ) {
      std::cerr << "too many OR definitions.\n";
      exit(-1);
    }

    p_or = &(or_array[or_count]);
    csrc >> i;
    if ( i < 0 ) {
      break;
    } else if ( i = 99 ) {
      p_or->p_bound = NULL;
    } else if ( i >= prim_count ) {
      std::cerr << "illegal primitive used in OR section.\n";
      exit(-1);
    } else {
      p_or->p_bound = &(primitives[i]);
    }
    p_or->length = 0;

    while ( csrc >> c >> i, (i >= 0) ) {
      if ( i >= and_count ) {
	std::cerr << "illegal AND used in OR section.\n";
	exit(-1);
      }
      p_or->p_and[p_or->length++] = &(and_array[i]);
    }
  }
}

void	clear_all(void)
{
  for ( int i = 0; i < prim_count; i++ ) { 
    if ( primitives[i].cube_flag == FALSE ) {
      primitives[i].form3D.p_quad->~quadratic3D();
      primitives[i].form3D.p_quad = NULL;
    } else {
      primitives[i].form3D.p_cube->~cube();
      primitives[i].form3D.p_cube = NULL;
      primitives[i].form2D.p_rect->~rectangle();
      primitives[i].form2D.p_rect = NULL;
    }
  }
}

// End Of File	---------------------------------------------------------------
